/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// experiments with the SainSmart DDS140 USB scope

#define _CRT_SECURE_NO_WARNINGS

#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>

#define THREAD 0

#ifdef _WIN32
#include <windows.h>
#include <gl\glew.h>
#include <gl\glu.h>

#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glu32.lib")
#pragma comment(lib, "glew32.lib")
#pragma comment(lib, "glew32s.lib")

#include "libusb.h"
#pragma comment(lib, "libusb-1.0.lib")
#else
#define LINUX
#endif

#ifdef LINUX
#include <libusb-1.0/libusb.h>
#include <unistd.h>

#include <GL/glx.h>
#include <GL/gl.h>
#include <GL/glu.h>

#endif

#include <wx/wx.h>
#include <wx/thread.h>
#include <wx/glcanvas.h>

class scope_frame;
scope_frame *frame = NULL;

class scope_usb
{
public:
    scope_usb() { scope_handle = NULL; }
    bool init_usb();
    bool usb_get_scope();
    int  scope_request(int req, int val);
    struct libusb_device_handle *scope_handle;
    int  scope_get_buffer(unsigned char* buf, size_t buf_size);
};

#define BUF_SIZE 0x20000
#define BUF_MASK 0x1ffff
#define BUF_TO_640 (BUF_SIZE/640)

class scope_render
{
public:
    scope_render() {};
    ~scope_render() {};
    scope_usb usb;
    
    void scope_init();
    int  scope_acquire();
    void render();
    void scope_capture_mode(int m, int a, int b, int gain_a, int gain_b);
    
    void setup_opengl_objects();
    void create_texture(int width, int height);
    void upload_texture(int width, int height, unsigned short* tex);
    void prepare_frame(int width, int height);
    GLuint scope_texture_id;
    GLuint scope_screen_quad;
    GLuint scope_shader_program;
    GLuint scope_attr_vertex;
    GLuint scope_unif_tex;
    bool ch2_enabled;
    unsigned short screen_buff[640 * 480];  // 16 bits, 5 6 5 format
    unsigned char scope_data[BUF_SIZE];
};

scope_render scope;
class scope_frame;

#if (THREAD)
class MyThread : public wxThread
{
public:
    MyThread(scope_frame *handler) : wxThread(wxTHREAD_DETACHED)
    {
        m_pHandler = handler;
    }
    ~MyThread();
protected:
    virtual ExitCode Entry();
    scope_frame *m_pHandler;
};
#endif

class scope_app : public wxApp
{
public:
    bool OnInit();
    int  scope_on_exit();
    void scope_on_idle(wxIdleEvent& evt);
};
class scope_frame : public wxFrame
{
public:
    scope_frame();
    ~scope_frame();

    wxGLCanvas* glPane;
    wxMenuBar*  menuBar;
    wxMenu*     fileMenu;
    wxComboBox* comb_speed;
    wxRadioBox* rb_ch1_acdc;
    wxRadioBox* rb_ch2_acdc;
    wxComboBox* comb_ch1_scale;
    wxComboBox* comb_ch2_scale;
    wxButton*   button_reconnect;
    wxButton*   button_send_go;
    void scope_on_about(wxCommandEvent& event);
    void scope_on_exit(wxCommandEvent& event);
    void scope_on_combo(wxCommandEvent& event);
    void scope_on_button_connect(wxCommandEvent& event);
    void scope_on_button_go(wxCommandEvent& event);
#if (THREAD)
protected:
    MyThread *m_pThread;
    wxCriticalSection m_pThreadCS; // protects the m_pThread pointer
    friend class MyThread; // allow it to access our m_pThread
#endif
    DECLARE_EVENT_TABLE()
};

#if (THREAD)
wxThread::ExitCode MyThread::Entry()
{
    static int i = 0;
    static int bytes = 0;
    while (!TestDestroy())
    {
        bytes += scope.scope_acquire();
        printf("thread %i %iMiB\n", i++, bytes / 1024 / 1024);
    }
    return (wxThread::ExitCode)0; // success
}
MyThread::~MyThread()
{
    wxCriticalSectionLocker enter(m_pHandler->m_pThreadCS);
    m_pHandler->m_pThread = NULL;
}
#endif

scope_frame::scope_frame() : wxFrame((wxFrame *)NULL, -1, wxT("scope"), wxDefaultPosition, wxSize(640 + 320, 480 + 120))
{
    this->CreateStatusBar(3);
    this->SetStatusText(_T("Scope"));
    this->Show(true);
    wxMenuBar* menubar = new wxMenuBar();
    wxMenu* file_menu = new wxMenu();
    file_menu->Append(wxID_EXIT, _T("&Quit"));
    menubar->Append(file_menu, _T("&File"));
    wxMenu* help_menu = new wxMenu();
    help_menu->Append(wxID_ABOUT, _T("&About"));
    menubar->Append(help_menu, _T("&Help"));
    this->SetMenuBar(menubar);

    // setup frame with GL output on left, controls on right
    wxBoxSizer* sizer = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer* sizer2 = new wxBoxSizer(wxVERTICAL);  // vertical list of controls

    int args[] = { WX_GL_RGBA, WX_GL_DOUBLEBUFFER, WX_GL_DEPTH_SIZE, 16, 0 };
    glPane = new wxGLCanvas(this, wxID_ANY, args, wxDefaultPosition, wxSize(640, 480), 0, wxT("GLCanvas"));
    wxGLContext* m_context = new wxGLContext(glPane);
    glPane->SetCurrent(*m_context);
#ifdef _WIN32
    GLenum err = glewInit();
    if (err) abort();
#endif

    sizer->Add(glPane, wxSizerFlags().FixedMinSize().Border(wxALL, 4));
    sizer->Add(sizer2, wxSizerFlags().Expand().Border(wxALL, 4));

    comb_speed = new wxComboBox(this, -1, wxT("100Mhz"), wxDefaultPosition, wxDefaultSize);
    comb_speed->Append(wxT("100Mhz"));
    comb_speed->Append(wxT("200Mhz (Ch1)"));
    comb_speed->Append(wxT("10Mhz"));
    comb_speed->Append(wxT("625Khz"));
    comb_speed->Append(wxT("39Khz"));
 //   comb_speed->Append(wxT("xxxKhz"));
  //  comb_speed->Append(wxT("yyyKhz"));

    wxArrayString a;    a.Add(wxT("Ch1 DC"));    a.Add(wxT("Ch1 AC"));
    rb_ch1_acdc = new wxRadioBox(this, -1, wxT("Ch1"), wxDefaultPosition, wxDefaultSize, a, 0);
    wxArrayString b;    b.Add(wxT("Ch2 DC"));    b.Add(wxT("Ch2 AC"));
    rb_ch2_acdc = new wxRadioBox(this, -1, wxT("Ch2"), wxDefaultPosition, wxDefaultSize, b, 0);

    comb_ch1_scale = new wxComboBox(this, -1, wxT("Ch1 5V"), wxDefaultPosition, wxDefaultSize);
    comb_ch1_scale->Append(wxT("Ch1 5v"));
    comb_ch1_scale->Append(wxT("Ch1 0.5v"));
    comb_ch1_scale->Append(wxT("Ch1 200mv"));
    comb_ch1_scale->Append(wxT("Ch1 100mv"));
    comb_ch1_scale->Append(wxT("Ch1 50mv"));

    comb_ch2_scale = new wxComboBox(this, -1, wxT("Ch2 5V"), wxDefaultPosition, wxDefaultSize);
    comb_ch2_scale->Append(wxT("Ch2 5v"));
    comb_ch2_scale->Append(wxT("Ch2 0.5v"));
    comb_ch2_scale->Append(wxT("Ch2 200mv"));
    comb_ch2_scale->Append(wxT("Ch2 100mv"));
    comb_ch2_scale->Append(wxT("Ch2 50mv"));

    button_reconnect=new wxButton(this, 4001, wxT("Reconnect USB"));
//    button_send_go=new wxButton(this, 4002, wxT("Send Go code"));

    sizer2->Add(comb_speed);
    sizer2->Add(rb_ch1_acdc);
    sizer2->Add(comb_ch1_scale);
    sizer2->Add(rb_ch2_acdc);
    sizer2->Add(comb_ch2_scale);
    sizer2->Add(button_reconnect);
 //   sizer2->Add(button_send_go);
    this->SetSizer(sizer);
    this->SetAutoLayout(true);
    this->Layout();
    scope.setup_opengl_objects();
    scope.create_texture(640, 480);
    //
#if (THREAD)
    m_pThread = new MyThread(this);
    if (m_pThread->Run() != wxTHREAD_NO_ERROR)
    {
        wxLogError(wxT("Can't create the thread!"));
        delete m_pThread;
        m_pThread = NULL;
    }
#endif
}
scope_frame::~scope_frame() {  }
BEGIN_EVENT_TABLE(scope_frame, wxFrame)
    EVT_MENU(wxID_EXIT, scope_frame::scope_on_exit)
    EVT_MENU(wxID_ABOUT, scope_frame::scope_on_about)
    EVT_COMBOBOX(wxID_ANY, scope_frame::scope_on_combo)
    EVT_RADIOBOX(wxID_ANY, scope_frame::scope_on_combo)
    EVT_BUTTON(4001, scope_frame::scope_on_button_connect)
    EVT_BUTTON(4002, scope_frame::scope_on_button_go)
 END_EVENT_TABLE()

DECLARE_APP(scope_app)
IMPLEMENT_APP(scope_app)

void scope_frame::scope_on_about(wxCommandEvent& event)
{
    wxMessageDialog aboutDialog(this, _T("donut6"), _T("About Scope"), wxOK );
    aboutDialog.ShowModal();
}
void scope_frame::scope_on_exit(wxCommandEvent& event)
{
    Close(TRUE);
}
void scope_frame::scope_on_combo(wxCommandEvent& event)
{
    int speed = comb_speed->GetCurrentSelection();
    int ch1_acdc = rb_ch1_acdc->GetSelection();
    int ch2_acdc = rb_ch2_acdc->GetSelection();
    int ch1_gain = comb_ch1_scale->GetCurrentSelection();
    int ch2_gain = comb_ch2_scale->GetCurrentSelection();
    if (speed < 0) speed = 0;
    if (ch1_acdc < 0) ch1_acdc = 0;
    if (ch2_acdc < 0) ch2_acdc = 0;
    if (ch1_gain < 0) ch1_gain = 0;
    if (ch2_gain < 0) ch2_gain = 0;
    scope.scope_capture_mode(speed, ch1_acdc, ch2_acdc, ch1_gain, ch2_gain);
}
void scope_frame::scope_on_button_connect(wxCommandEvent& event)
{
    scope.usb.usb_get_scope();
}
void scope_frame::scope_on_button_go(wxCommandEvent& event)
{
    scope.usb.scope_request(0x33, 0x00);   // "go" code
}

void scope_app::scope_on_idle(wxIdleEvent& evt)
{
#if (!THREAD)
    scope.scope_acquire();
#endif
    scope.render();
    evt.RequestMore();
}
bool scope_app::OnInit()
{
#ifdef _WIN32
    AllocConsole();
    SetConsoleTitle(L"scope");
    FILE *blah;
    freopen_s(&blah, "CONOUT$", "w", stdout);
    freopen_s(&blah, "CONOUT$", "w", stderr);
#endif
    scope.scope_init();
    frame = new scope_frame();
    SetTopWindow(frame);
    Connect(wxID_ANY, wxEVT_IDLE, wxIdleEventHandler(scope_app::scope_on_idle));
    return true;
}
int  scope_app::scope_on_exit()
{
    return 0;
}

void scope_render::setup_opengl_objects()
{
    static const GLfloat scope_vertex_data[] = {
        -1.0, -1.0, 1.0, 1.0,
        1.0, -1.0, 1.0, 1.0,
        -1.0, 1.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0
    };
    const char *scope_vert_source =
    "attribute vec4 vertex;"
    "varying vec2 tcoord;"
    "void main(void) {"
    " vec4 pos = vertex;"
    " gl_Position = pos;"
    " tcoord.x = (vertex.x* 0.5)+0.5;"
    " tcoord.y = (vertex.y*-0.5)+0.5;"
    "}";
    
    const char *scope_frag_source =
    "varying vec2 tcoord;"
    "uniform sampler2D tex;"
    "void main(void) {"
    "  gl_FragColor = texture2D(tex, tcoord);"
    "}";
    GLuint vshader;
    GLuint fshader;
    vshader = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(vshader, 1, &scope_vert_source, 0);
    glCompileShader(vshader);
    
    fshader = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(fshader, 1, &scope_frag_source, 0);
    glCompileShader(fshader);
    
    scope_shader_program = glCreateProgram();
    glAttachShader(scope_shader_program, vshader);
    glAttachShader(scope_shader_program, fshader);
    glLinkProgram(scope_shader_program);
    
    scope_attr_vertex = glGetAttribLocation(scope_shader_program, "vertex");
    scope_unif_tex = glGetUniformLocation(scope_shader_program, "tex");
    
    glGenBuffers(1, &scope_screen_quad);
    glBindBuffer(GL_ARRAY_BUFFER, scope_screen_quad);
    glBufferData(GL_ARRAY_BUFFER, sizeof(scope_vertex_data), scope_vertex_data, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}
void scope_render::prepare_frame(int width, int height)
{
    glViewport(0, 0, width, height);
    glBindFramebuffer(GL_FRAMEBUFFER,0);
}
void scope_render::create_texture(int width, int height)
{
    glActiveTexture(GL_TEXTURE0);
    printf("creating texture... %i x %i\n", width, height);
    glGenTextures(1, &scope_texture_id);
    glBindTexture(GL_TEXTURE_2D, scope_texture_id);

    glPixelStorei(GL_UNPACK_ALIGNMENT, 2);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_SHORT_5_6_5, 0);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
}
void scope_render::upload_texture(int width, int height, unsigned short* tex)
{
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, scope_texture_id);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 2);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_RGB, GL_UNSIGNED_SHORT_5_6_5, tex);
}
void scope_render::render()
{
    prepare_frame(640, 480);
    for(int a=0; a<(640*480); a++ )
    {
#if 0
        int fade=(screen_buff[a]>>5) & 0x3f;
        if (fade) fade-=5; if (fade<0) fade=0;
        screen_buff[a]=(fade) << 5;
#else
        screen_buff[a] = 0;
#endif
    }
    // display ch1 and ch2...
    for (int s = 0; s<BUF_MASK; s += 0x1)
    {
        int dy = 240 - (scope_data[(s&(BUF_MASK - 1))] - 0x80);
        if (dy<0) dy = 0; if (dy >= 480) dy = 479;
        int dx = s / BUF_TO_640; if (dx > 639) dx = 639;
        screen_buff[dx + (dy * 640)] = 0x3f << 5;    // green. 16 bits, 5 6 5 format
        if (ch2_enabled)
        {
           dy = 340 - (scope_data[(s & BUF_MASK) | 1] - 0x80);;
           if (dy < 0) dy = 0; if (dy >= 480) dy = 479;
           screen_buff[dx + (dy * 640)] = 0x1f << 11;    // red
        }
    }
    upload_texture(640, 480, screen_buff);

    glBindBuffer(GL_ARRAY_BUFFER, scope_screen_quad);
    glUseProgram(scope_shader_program);
    glVertexAttribPointer(scope_attr_vertex, 4, GL_FLOAT, 0, 16, 0);
    glEnableVertexAttribArray(scope_attr_vertex);                       // shader
    //  glEnable(GL_TEXTURE_2D);
    glUniform1i(scope_unif_tex, 0);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    glBindBuffer(GL_ARRAY_BUFFER, 0);       // done here
    glUseProgram(0);

    frame->glPane->SwapBuffers();
}
int  scope_render::scope_acquire()
{
    int w = 0;
    int resync = 0;
    usb.scope_request(0x33, 0x00);   // "go" code
    // check FIFO status:
    while ((w != 0x2c) && (w != 0x21))
    {
        w = usb.scope_request(0x50, 0);
        resync++;
        if (resync>1000) break;
    }
    while (w != 0x21)
    {
        w = usb.scope_request(0x50, 0);
        resync++;
        if (resync>2000) break;
    }
    if (w == 0x21) // fifo full
    {
        int r = usb.scope_get_buffer(scope_data, sizeof(scope_data));
        if (r < 0) return 0;
        return r;
    }
    return 0;
}
void scope_render::scope_capture_mode(int clock_mode, int ch1_acdc, int ch2_acdc, int gain_a, int gain_b)
{
    int CH1_ACDC = 0x10;           // DC
    int CH2_ACDC = 0x10;           // DC
    if (ch1_acdc) CH1_ACDC = 0x0;  // AC
    if (ch2_acdc) CH2_ACDC = 0x0;  // AC

    //  Command  Bit Field ..... Description
    //    0x22   0000 0aa0    
    //                 00  ..... Gain Ch1   (0x0) port B
    //                 01  ..... Gain Ch1   (0x2)
    //                 10  ..... Gain Ch1   (0x4)
    //    0x23   0b00 0000 
    //            0        ..... Gain Ch2   (0x00) port E
    //            1        ..... Gain Ch2   (0x40)
    //    0x24   000c abbr
    //                   0 ..... Relay
    //                   1 ..... Relay - Ch1 to (Ch1 & Ch2) ADC   
    //                 00  ..... Gain Ch2   (| 0x0)
    //                 01  ..... Gain Ch2   (| 0x2)
    //                 10  ..... Gain Ch2   (| 0x4)
    //                 11  ..... Gain Ch2   (| 0x6)
    //                0    ..... Gain Ch1   (| 0x0)
    //                1    ..... Gain Ch1   (| 0x8)
    //              0      ..... Ch2 AC     (| 0x00)
    //              1      ..... Ch2 DC     (| 0x10)
    //    0x94   000d ssss
    //              0      ..... Ch1 AC
    //              1      ..... Ch1 DC
    //                0000 ..... 100Mhz  (0x0)
    //                1100 .....  10Mhz  (0xc)  80/0x8
    //                1000 ..... 625Khz  (0x8)  80/0x80
    //                1011 .....  39Khz  (0xb)  80/0x800
    //                1010 .....   ?     (0xa)

    int reg22 = 0;
    int reg23 = 0;
    int reg94_clock = 0;
    int reg24_a = 0;
    int reg24_b = 0;
    switch (gain_a)
    {
        case 0:  // Ch1 5v range  (aaa=001)
            reg22=0x00;  reg24_a = 0x08;
            break;
        case 1: // Ch1 0.5v range (aaa=011)
            reg22=0x02;  reg24_a = 0x08;
            break;
        case 2:  // Ch1 200mv range (aaa=000)
            reg22=0x00;  reg24_a = 0x00;
            break;
        case 3:  // Ch1 100mv range (aaa=010)
            reg22=0x02;  reg24_a = 0x00;
            break;
        case 4:   // Ch1 50mv range (aaa=100)
            reg22=0x04;  reg24_a = 0x00;
            break;
    }

    switch (gain_b)
    {
        case 0: // Ch2 5v range  (bbb=000)
            reg23=0x00; reg24_b = 0x00;
            break;
        case 1: // Ch2 0.5v range  (bbb=001)
            reg23 = 0x00; reg24_b = 0x02;
            break;
        case 2: // Ch2 200mv range  (bbb=100)
            reg23 = 0x40; reg24_b = 0x00;
            break;
        case 3: // Ch2 100mv range (bbb=101)
            reg23 = 0x40; reg24_b = 0x02;
            break;
        case 4: //  Ch2 50mv range (bbb=110)
            reg23 = 0x40;  reg24_b = 0x04;
            break;
    }
    ch2_enabled = true;
    switch (clock_mode)
    {
    case 0:
        reg94_clock = 0x0;        // 100Mhz
        break;
    case 1:
        reg24_a = 0x9;            // probe 1 to both ADCs (200 Mhz)
        reg24_b = 0;
        reg94_clock = 0x0;        // 100Mhz
        ch2_enabled = false;
        break;
    case 2:
        reg94_clock = 0xc;        // 10Mhz
        break;
    case 3:
        reg94_clock = 0x8;        // 625Khz
        break;
    case 4:
        reg94_clock = 0xb;        // 39Khz
        break;
    case 5:
        reg94_clock = 0xa;        // ?Mhz
        break;
    case 6:
        reg94_clock = 0xd;        // ?Mhz
        break;
    default:
        reg94_clock = 0x0;        // 100Mhz
    }
    usb.scope_request(0x22, reg22);
    usb.scope_request(0x23, reg23);
    usb.scope_request(0x24, reg24_a | reg24_b | CH2_ACDC);
    usb.scope_request(0x94, reg94_clock | CH1_ACDC);
}
void scope_render::scope_init()
{
    usb.init_usb();

    usb.scope_request(0x76, 0xe8);  // timer for square wave output?
    usb.scope_request(0x77, 0x9b);  // timer  
    usb.scope_request(0x78, 0xe8);  // timer 
    usb.scope_request(0x79, 0x9b);  // timer 


    usb.scope_request(0x63, 0x04);
    usb.scope_request(0x75, 0x00);
    usb.scope_request(0x34, 0x00);  // 0=enable scope mode 1=enable logic analyzer mode

 //   usb.scope_request(0x7a, 0xfb);  // timer 
 //   usb.scope_request(0x7b, 0x8c);  // 
 //   usb.scope_request(0x7c, 0xff);  // 
 //   usb.scope_request(0x7d, 0xc4);  // 

    scope_capture_mode(0, 0, 0, 0, 0); // 100Mhz, 2 probes

    usb.scope_request(0xe7, 0x00); 
}

bool scope_usb::init_usb()
{
    int r = libusb_init(NULL);
    if (r < 0)  return false;
    return usb_get_scope();
}
bool scope_usb::usb_get_scope()
{
    if (scope_handle)
        libusb_close(scope_handle);
    printf("Looking for scope device 8312:8312\n");
    scope_handle = libusb_open_device_with_vid_pid(NULL, 0x8312, 0x8312);
    if (scope_handle == 0)
    {
        fprintf(stderr, "Could not open device\n");
        return false;
    }

    if (scope_handle)
    {
        int r = libusb_claim_interface(scope_handle, 0);
        if (r < 0)
        {
            fprintf(stderr, "libusb_claim_interface error %d\n", r);
            return false;
        }
        return true;
    }
    return false;
}
int  scope_usb::scope_request(int req, int val)
{
    unsigned char data = 0;
    if (scope_handle)
        int r = libusb_control_transfer(scope_handle, 0 | LIBUSB_ENDPOINT_IN, req, val, 0, &data, 1, 0);
    return data;
}
int  scope_usb::scope_get_buffer(unsigned char* buf, size_t buf_size)
{
    int r;
    int transferred = 0;
    if (scope_handle)
    {
        r = libusb_bulk_transfer(scope_handle, 0x2 | LIBUSB_ENDPOINT_IN, buf, buf_size, &transferred, 500);
        if (r < 0)
        {
            fprintf(stderr, "scope data read error %d\n", r);
            return r;
        }
    }
    return transferred;
}

